#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/stat.h>

#include <TH/th_fnt.h>
#include <GFX/tga.h>
#include <REAP/reap_string.h>

// Calculate number of bytes needed to hold so many bits
int bitbytes(int bits)
{
    int bytes = bits + 7;
    bytes &= ~7;
    bytes >>= 3;

    return bytes;
}

int main(int argc, char ** argv)
{
    sTGAHeader  t;
    sFNTHeader  fh;
    sFNTData    fd;
    sCharacterInfo * chars;
    sRGBA * colors;
    int i1, i2;
    //unsigned char * data;
    unsigned char * FntTextureData;

    char  out_char_name[512];
    FILE * out_char_file;

    FILE *          FntFile;
    char *          FntFileName;
    unsigned long   FntFileLength;

    FILE *          TgaFile;
    char *          TgaFileName;

    FILE *          TextFile;   // File containing extra information about font
    char *          TextFileName;

    char *          DirectoryName;
    char *          DirectoryName2;

    printf("FNT2TGA for Tony Hawk's Underground by reapher\n");
    printf("**********************************************\n");

    if(argc < 2)
    {
        printf("Enter the names of FNT files to convert to TGA when calling program\n");
    }
    else
    {
        for(i1 = 1; i1 < argc; i1++)
        {
            char * tmp_ptr;

            printf("Input: %s\n", argv[i1]);
            FntFile = fopen(argv[i1], "rb");

            FntFileName     = cpy_str_after_chr(argv[i1], '\\', -1);

            DirectoryName   = cpy_str_till_chr(argv[i1], '.', -1);
            DirectoryName2  = cpy_str_after_chr(DirectoryName, '\\', -1);

            TextFileName    = cpy_str_after_chr(argv[i1], '\\', -1);
            tmp_ptr = strrchr(TextFileName, '.');
            *(tmp_ptr + 1) = 't';
            *(tmp_ptr + 2) = 'x';
            *(tmp_ptr + 3) = 't';

            TgaFileName     = cpy_str_after_chr(argv[i1], '\\', -1);
            tmp_ptr = strrchr(TgaFileName, '.');
            *(tmp_ptr + 1) = 't';
            *(tmp_ptr + 2) = 'g';
            *(tmp_ptr + 3) = 'a';

            if(FntFile)
            {
                fseek(FntFile, 0, SEEK_END);
                FntFileLength = ftell(FntFile);
                fseek(FntFile, 0, SEEK_SET);

                if(FntFileLength > sizeof(fh))
                {
                    fread(&fh, sizeof(fh), 1, FntFile);
                    if(fh.NumberOfChars)
                    {
                        chars = malloc(sizeof(sCharacterInfo) * fh.NumberOfChars);

                        int i3;

                        for(i3 = 0; i3 < fh.NumberOfChars; i3++)
                        {
                            fread(&chars[i3].Base, 2, 1, FntFile);
                            fread(&chars[i3].Ascii, 2, 1, FntFile);
                        }

                        fread(&fd.DataLength, 4, 1, FntFile);
                        fread(&fd.Width, 2, 1, FntFile);
                        fread(&fd.Height, 2, 1, FntFile);
                        fread(&fd.Depth, 2, 1, FntFile);
                        fread(&fd.Extra, 6, 1, FntFile);

                        // Malloc enough space for pixel data
                        int pixel_data_size = fd.Width * fd.Height * bitbytes(fd.Depth);
                        FntTextureData = malloc(pixel_data_size);

                        fread(FntTextureData, pixel_data_size, 1, FntFile);

                        int file_pos = ftell(FntFile);

                        if((fh.FileLength - file_pos) < ((fh.NumberOfChars * 8) + 1024))
                        {
                            printf("--Invalid color lookup table!\n");
                        }
                        else
                        {
                            int i4;

                            colors = malloc(1024);
                            fread(colors, 4, 256, FntFile);

                            int num_subtextures;

                            fread(&num_subtextures, 4, 1, FntFile);

                            printf("--Number of sub-textures: %d\n", num_subtextures);

                            for(i4 = 0; i4 < fh.NumberOfChars; i4++)
                            {
                                fread(&chars[i4].PosX, 2, 1, FntFile);
                                fread(&chars[i4].PosY, 2, 1, FntFile);
                                fread(&chars[i4].Width, 2, 1, FntFile);
                                fread(&chars[i4].Height, 2, 1, FntFile);
                            }

                            file_pos = ftell(FntFile);

                            if(file_pos < fh.FileLength)
                            {
                                printf("--Extra data in FNT file!\n");
                            }

                            memset(&t, 0, sizeof(t));

                            t.ImageType = 2;
                            t.Width = fd.Width;
                            t.Height = fd.Height;
                            t.Depth = 32;
                            t.ImageDescriptor = 32; // Specify image vertically upside down

                            TgaFile = fopen(TgaFileName, "wb");

                            fwrite(&t.ImageIdLength, 1, 1, TgaFile);
                            fwrite(&t.ColorMapType, 1, 1, TgaFile);
                            fwrite(&t.ImageType, 1, 1, TgaFile);
                            fwrite(&t.CMFirstEntryIndex, 1, 2, TgaFile);
                            fwrite(&t.CMLength, 1, 2, TgaFile);
                            fwrite(&t.CMEntrySize, 1, 1, TgaFile);
                            fwrite(&t.XOrigin, 1, 2, TgaFile);
                            fwrite(&t.YOrigin, 1, 2, TgaFile);
                            fwrite(&t.Width, 1, 2, TgaFile);
                            fwrite(&t.Height, 1, 2, TgaFile);
                            fwrite(&t.Depth, 1, 1, TgaFile);
                            fwrite(&t.ImageDescriptor, 1, 1, TgaFile);

                            for(i2 = 0; i2 < pixel_data_size; i2++)
                            {
                                fwrite(&colors[FntTextureData[i2]].R, 1, 1, TgaFile);
                                fwrite(&colors[FntTextureData[i2]].G, 1, 1, TgaFile);
                                fwrite(&colors[FntTextureData[i2]].B, 1, 1, TgaFile);
                                fwrite(&colors[FntTextureData[i2]].A, 1, 1, TgaFile);
                            }

                            fclose(TgaFile);

                            TextFile = fopen(TextFileName, "w");

                            mkdir(DirectoryName2, S_IRUSR | S_IWUSR | S_IXUSR);

                            fprintf(TextFile, "FNT %s\n", FntFileName);

                            for(i4 = 0; i4 < fh.NumberOfChars; i4++)
                            {
                                chars[i4].Ascii &= 0xFF;
                                sTGAHeader char_header;

                                memset(&char_header, 0, sizeof(char_header));
                                char_header.ImageType = 2;
                                char_header.Width = chars[i4].Width;
                                char_header.Height = chars[i4].Height;
                                char_header.Depth = 32;
                                char_header.ImageDescriptor = 32;

                                memset(out_char_name, 0, 512);
                                sprintf(out_char_name, "%s\\%02X.tga", DirectoryName2, chars[i4].Ascii);

                                out_char_file = fopen(out_char_name, "wb");

                                fwrite(&char_header.ImageIdLength, 1, 1, out_char_file);
                                fwrite(&char_header.ColorMapType, 1, 1, out_char_file);
                                fwrite(&char_header.ImageType, 1, 1, out_char_file);
                                fwrite(&char_header.CMFirstEntryIndex, 1, 2, out_char_file);
                                fwrite(&char_header.CMLength, 1, 2, out_char_file);
                                fwrite(&char_header.CMEntrySize, 1, 1, out_char_file);
                                fwrite(&char_header.XOrigin, 1, 2, out_char_file);
                                fwrite(&char_header.YOrigin, 1, 2, out_char_file);
                                fwrite(&char_header.Width, 1, 2, out_char_file);
                                fwrite(&char_header.Height, 1, 2, out_char_file);
                                fwrite(&char_header.Depth, 1, 1, out_char_file);
                                fwrite(&char_header.ImageDescriptor, 1, 1, out_char_file);

                                int i6 = 0, i7 = 0;

                                for(i6 = chars[i4].PosY; i6 < (chars[i4].PosY + chars[i4].Height); i6++)
                                {
                                    for(i7 = chars[i4].PosX; i7 < (chars[i4].PosX + chars[i4].Width); i7++)
                                    {
                                        fwrite(&colors[FntTextureData[(i6 * t.Width) + i7]].R, 1, 1, out_char_file);
                                        fwrite(&colors[FntTextureData[(i6 * t.Width) + i7]].G, 1, 1, out_char_file);
                                        fwrite(&colors[FntTextureData[(i6 * t.Width) + i7]].B, 1, 1, out_char_file);
                                        fwrite(&colors[FntTextureData[(i6 * t.Width) + i7]].A, 1, 1, out_char_file);
                                    }
                                }

                                fclose(out_char_file);

                                fprintf(TextFile, "CHAR %c %s\n", chars[i4].Ascii, out_char_name);
                            }

                            fclose(TextFile);

                            //if(data) free(data);

                            free(colors);
                        }

                        free(FntTextureData);

                        free(chars);

                        free(FntFileName);
                        free(TextFileName);
                        free(DirectoryName);
                        free(DirectoryName2);
                        free(TgaFileName);
                    }
                    else
                    {
                        printf("No characters in FNT file!\n");
                    }
                }
                else
                {
                    printf("\tUnable to process file!\n");
                }
            }

            fclose(FntFile);
        }
    }

    system("PAUSE");

    return 0;
}
